from asyncio import Lock
from logging import Logger
from typing import Set, Optional


class OptimizationQueue:
    """
    Responsible to properly control what is being processed
    """

    def __init__(self, queued_pids: Set[int], logger: Optional[Logger] = None):
        self._queued_pids = queued_pids
        self._lock_queued_pids = Lock()
        self._logger = logger

    def _log(self, msg: str):
        if self._logger:
            self._logger.debug(msg)

    async def add_pid(self, pid: int) -> bool:
        if pid is not None:
            async with self._lock_queued_pids:
                if pid not in self._queued_pids:
                    self._log(f"Adding pid {pid} to the optimization queue")
                    self._queued_pids.add(pid)
                    return True

        return False

    async def remove_pids(self, *pids: int):
        if pids is not None:
            async with self._lock_queued_pids:
                for pid in pids:
                    if pid is not None:
                        self._log(f"Removing pid {pid} from the optimization queue")
                        self._queued_pids.discard(pid)

    def get_view(self) -> Optional[Set[int]]:
        if self._queued_pids is not None:
            return {*self._queued_pids}

    @classmethod
    def empty(cls) -> "OptimizationQueue":
        return OptimizationQueue(set())



import os
from logging import Logger
from typing import Optional, List, Dict, Set

from guapow.common.model import ScriptSettings
from guapow.common.system import run_async_process, ProcessTimedOutError
from guapow.common.users import is_root_user


class RunScripts:

    def __init__(self, name: str, root_allowed: bool, logger: Logger):
        self._name = name
        self._log = logger
        self.root_allowed = root_allowed

    @staticmethod
    def get_environ(env: Optional[Dict[str, str]] = None) -> Dict[str, str]:
        final_env = env if env else dict(os.environ)

        if 'DISPLAY' not in final_env:
            final_env['DISPLAY'] = ':0'

        return final_env

    async def _execute_scripts(self, settings: ScriptSettings, user_id: Optional[int] = None,
                               user_env: Optional[Dict[str, str]] = None) -> Set[int]:
        pids = set()
        env = self.get_environ(user_env)

        valid_timeout = settings.has_valid_timeout()

        if not valid_timeout and settings.timeout is not None:
            self._log.warning(f"Invalid {self._name} scripts timeout defined: {settings.timeout}. "
                              f"No script will be awaited")

        should_wait = settings.wait_execution or valid_timeout

        for cmd in settings.scripts:
            self._log.info(f"{'Waiting' if should_wait else 'Starting'} {self._name} script: {cmd}")
            try:
                pid, _, output = await run_async_process(cmd=cmd, user_id=user_id, custom_env=env,
                                                         wait=settings.wait_execution, timeout=settings.timeout,
                                                         output=False)

                if pid is not None:
                    pids.add(pid)
                    if should_wait:
                        self._log.info(f"{self._name.capitalize()} script finished: {cmd} (pid={pid})")
                else:
                    err_output = f": {output}" if output else ""
                    self._log.error(f"Unexpected error when running {self._name} script '{cmd}'{err_output}")
            except ProcessTimedOutError as e:
                self._log.warning(f"{self._name.capitalize()} script '{cmd}' timed out (pid={e.pid})")

        return pids

    async def run(self, scripts: List[ScriptSettings], user_id: Optional[int],
                  user_env: Optional[Dict[str, str]]) -> Set[int]:
        current_user_id = os.getuid()
        root_user = is_root_user(current_user_id)

        pids = set()

        for settings in scripts:
            if root_user:
                if not settings.run_as_root and user_id is not None and not is_root_user(user_id):
                    pids.update(await self._execute_scripts(settings, user_id, user_env))
                elif self.root_allowed:
                    pids.update(await self._execute_scripts(settings))
                else:
                    self._log.warning(f"{self._name.capitalize()} scripts {settings.scripts} are not allowed "
                                      f"to run at the root level")
            elif settings.run_as_root:
                self._log.warning(f"Cannot execute {self._name} scripts {settings.scripts} as root user")
            elif user_id is None:
                pids.update(await self._execute_scripts(settings))
            elif current_user_id == user_id:
                pids.update(await self._execute_scripts(settings, user_env=user_env))
            else:
                self._log.warning(f"Cannot execute {self._name} scripts {settings.scripts} as user {user_id}")

        return pids



import asyncio
import os
import time
from asyncio import Task
from datetime import timedelta, datetime
from typing import Optional, Iterable, List

from guapow.common.dto import OptimizationRequest
from guapow.common.profile import get_possible_profile_paths_by_priority, \
    get_default_profile_name
from guapow.common.system import map_processes_by_parent, find_process_children
from guapow.service.optimizer.launcher import LauncherMapperManager
from guapow.service.optimizer.profile import OptimizationProfile, OptimizationProfileReader
from guapow.service.optimizer.task.manager import TasksManager, run_tasks
from guapow.service.optimizer.task.model import OptimizationContext, OptimizedProcess
from guapow.service.optimizer.watch import DeadProcessWatcherManager


class OptimizationHandler:

    def __init__(self, context: OptimizationContext, tasks_man: TasksManager, watcher_man: DeadProcessWatcherManager,
                 profile_reader: OptimizationProfileReader):
        super(OptimizationHandler, self)
        self.context = context
        self._queue = context.queue
        self._log = context.logger
        self._tasks_man = tasks_man
        self._watcher_man = watcher_man
        self._profile_reader = profile_reader
        self._launcher_mapper = LauncherMapperManager(check_time=context.launcher_mapping_timeout,
                                                      found_check_time=context.launcher_mapping_found_timeout,
                                                      logger=context.logger)
    @property
    def children_timeout(self):
        return self.context.search_children_timeout

    @property
    def children_found_timeout(self):
        return self.context.search_children_found_timeout

    async def _read_valid_profile(self, name: str, add_settings: Optional[str], user_id: Optional[int], user_name: Optional[str], request: OptimizationRequest) -> Optional[OptimizationProfile]:
        for file_path in get_possible_profile_paths_by_priority(name=name, user_id=user_id, user_name=user_name):
            if file_path:
                try:
                    return await self._profile_reader.read_valid(profile_path=file_path, add_settings=add_settings,
                                                                 handle_not_found=False)
                except FileNotFoundError:
                    self._log.debug(f"Profile file '{file_path}' not found (request={request.pid})")

    async def _load_valid_profile(self, request: OptimizationRequest) -> Optional[OptimizationProfile]:
        profile = None

        if request.profile:
            profile = await self._read_valid_profile(request.profile, request.profile_config, request.user_id, request.user_name, request)

            if profile:
                self._log.info(f"Valid profile '{profile.name}' ({profile.path}) found (request={request.pid})")

        if not profile:
            profile = await self._read_valid_profile(get_default_profile_name(), request.profile_config, request.user_id, request.user_name, request)

            if profile:
                pre_msg = "No existing/valid profile '{}'. ".format(request.profile) if request.profile else "Request has no profile defined. "
                self._log.warning(f"{pre_msg}Profile '{profile.path}' will be used instead (request={request.pid})")

        return profile

    def map_valid_config(self, config: str) -> Optional[OptimizationProfile]:
        if config:
            profile = self._profile_reader.map(profile_str=config)

            if profile and profile.is_valid():
                return profile

        self._log.warning("No optimization settings defined in configuration: {}".format(config.replace('\n', ' ')))

    async def _handle_process(self, process: OptimizedProcess, proc_tasks: Optional[Iterable[Task]] = None,
                              env_tasks: Optional[Iterable[Task]] = None, mapped: bool = False) -> None:

        await self._queue.add_pid(process.pid)

        if env_tasks:
            self._log.debug(f"Awaiting environment tasks required by process '{process.pid}'")
            await asyncio.gather(*run_tasks(env_tasks, process))

        should_be_watched = process.should_be_watched()

        if should_be_watched:
            self._log.debug(f"Process '{process.pid}' should be watched")
            await self._watcher_man.watch(process)

            if mapped:
                await self._queue.remove_pids(process.source_pid)

        if proc_tasks:
            self._log.debug(f"Awaiting process tasks required by the process '{process.pid}'")
            await asyncio.gather(*run_tasks(proc_tasks, process))

        if not should_be_watched:
            self._log.debug(f"Process '{process.pid}' does not require watching")
            related_pids = process.get_pids()

            if related_pids:
                self._log.debug(f"Disassociating process '{process.pid}' related pids "
                                f"({', '.join(str(p) for p in related_pids)}) from the optimization queue")
                await self._queue.remove_pids(*related_pids)

        request = process.request
        exec_time = time.time() - request.created_at
        self._log.debug(f"Optimization request for process '{process.pid}' took {exec_time:.4f} seconds"
                        f"{f' (source_pid={process.pid})' if request.pid != process.pid else ''}")

    async def optimize_children(self, target_parents: Iterable[OptimizedProcess], env_tasks: Iterable[Task],
                                proc_tasks: Iterable[Task]) -> None:
        already_found = set()  # children already found

        # max time period to end the iteration in case a result is found
        latest_found_timestamp: Optional[datetime] = None

        time_init = datetime.now()

        # max time period to search for children
        timeout = time_init + timedelta(seconds=self.children_timeout)

        while datetime.now() < timeout:
            if latest_found_timestamp and datetime.now() >= latest_found_timestamp:
                ppids = ", ".join(sorted(str(p.pid) for p in target_parents))
                self._log.debug(f"Children search timed out earlier (ppids={ppids})")
                return

            ppid_comm_pid = await map_processes_by_parent()

            for parent in target_parents:
                for pid_, comm_, ppid_ in find_process_children(ppid=parent.pid,
                                                                processes_by_parent=ppid_comm_pid,
                                                                already_found=already_found):
                    if self.children_found_timeout >= 0:
                        latest_found_timestamp = datetime.now() + timedelta(seconds=self.children_found_timeout)

                    self._log.info(f"Child process found: {comm_} (pid={pid_}, ppid={ppid_})")
                    child = parent.clone()
                    child.pid = pid_
                    await self._handle_process(child, proc_tasks, env_tasks)

        time_end = datetime.now()
        ppids = ", ".join(sorted(str(p.pid) for p in target_parents))
        self._log.debug(f"Children search timed out in {(time_end - time_init).total_seconds():.2f} (ppids={ppids})")

    async def handle(self, request: OptimizationRequest):
        request.prepare()

        if not os.path.exists(f'/proc/{request.pid}'):
            self._log.warning(f'Process {request.pid} does not exist. No optimization will be applied.')
            await self.context.queue.remove_pids(request.pid)
        else:
            if request.has_full_configuration():
                profile = self.map_valid_config(request.config)
            else:
                profile = await self._load_valid_profile(request)

            if not profile:
                self._log.warning(f"No optimizations available for process '{request.pid}'")

            source_process = OptimizedProcess(request=request, created_at=time.time(), profile=profile)

            env_tasks: Optional[Iterable[Task]] = None
            proc_tasks: Optional[Iterable[Task]] = None
            handled_processes: List[OptimizedProcess] = list()

            if profile:
                env_tasks = await self._tasks_man.get_available_environment_tasks(source_process)

                if source_process.profile.process:
                    proc_tasks = await self._tasks_man.get_available_process_tasks(source_process)

                async for pid in self._launcher_mapper.map_pids(source_process.request, source_process.profile):
                    if pid is not None and pid != source_process.pid:
                        mapped_process = source_process.clone()
                        mapped_process.pid = pid
                        await self._handle_process(mapped_process, proc_tasks, env_tasks, mapped=True)
                        handled_processes.append(mapped_process)

            if not handled_processes:
                # if no other process was handled yet, it means the source process was not mapped as another processes.
                # so it should be optimized
                await self._handle_process(source_process, proc_tasks, env_tasks)
                handled_processes.append(source_process)

            if handled_processes:
                if self.children_timeout <= 0:
                    ppids = ", ".join(sorted(str(p.pid) for p in handled_processes))
                    self._log.debug(f"Not looking for processes children (ppids={ppids})")
                else:
                    await self.optimize_children(handled_processes, proc_tasks, env_tasks)



import asyncio
import os
import subprocess
import traceback
from datetime import datetime, timedelta
from io import StringIO
from re import Pattern
from typing import Optional, Tuple, Dict, List, Set, Callable, TypeVar, Collection, Generator

BAD_USER_ENV_VARS = {'LD_PRELOAD'}
T = TypeVar('T')


class ProcessTimedOutError(Exception):
    def __init__(self, pid: int):
        super(ProcessTimedOutError, self).__init__()
        self.pid = pid


def syscall(cmd: str, shell: bool = True, cwd: Optional[str] = None, custom_env: Optional[dict] = None, stdin: bool = False, return_output: bool = True) -> Tuple[int, Optional[str]]:
    params = {
        'args': cmd.split(' ') if not shell else [cmd],
        'stdout': subprocess.PIPE if return_output else subprocess.DEVNULL,
        'stderr': subprocess.STDOUT if return_output else subprocess.DEVNULL,
        'shell': shell
    }

    if not stdin:
        params['stdin'] = subprocess.DEVNULL

    if cwd is not None:
        params['cwd'] = cwd

    if custom_env is not None:
        params['env'] = custom_env

    try:
        p = subprocess.Popen(**params)
    except:
        traceback.print_exc()
        return False, None

    str_output = StringIO('') if return_output else None

    if str_output:
        for stream in p.stdout:
            string = stream.decode()

            if return_output:
                str_output.write(string)

    p.wait()

    if str_output and return_output:
        str_output.seek(0)
        return p.returncode, str_output.read()
    else:
        return p.returncode, None


async def async_syscall(cmd: str, shell: bool = True, cwd: Optional[str] = None, custom_env: Optional[Dict[str, str]] = None, stdin: bool = False, return_output: bool = True, wait: bool = True) -> Tuple[int, Optional[str]]:
    params = {
        'cmd': cmd,
        'stdout': subprocess.PIPE if return_output else subprocess.DEVNULL,
        'stderr': subprocess.STDOUT if return_output else subprocess.DEVNULL,
    }

    if not stdin:
        params['stdin'] = subprocess.DEVNULL

    if cwd is not None:
        params['cwd'] = cwd

    if custom_env is not None:
        params['env'] = custom_env

    try:
        if shell:
            p = await asyncio.create_subprocess_shell(**params)
        else:
            p = await asyncio.create_subprocess_exec(**params)
    except:
        traceback.print_exc()
        return False, None

    output = None

    if return_output:
        string = StringIO()

        try:
            async for stream in p.stdout:
                if stream:
                    string.write(stream.decode())
        except ValueError:
            return 1, None

        string.seek(0)
        output = string.read()

    if wait:
        await p.wait()

    return p.returncode, (output if output is not None else None)


async def match_syscall(cmd: str, match: Callable[[str], Optional[T]]) -> Optional[T]:
    p = await asyncio.create_subprocess_shell(cmd=cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, stdin=subprocess.DEVNULL)

    try:
        async for stream in p.stdout:
            if stream:
                res = match(stream.decode())

                if res is not None:
                    return res

    except ValueError:
        return


async def find_process_by_name(name_pattern: Pattern, last_match: bool = False) -> Optional[Tuple[int, str]]:
    def _match(line: str) -> Optional[Tuple[int, str]]:
        line_strip = line.strip()

        if line_strip:
            line_split = line_strip.split('#', 1)

            if len(line_split) > 1:
                name = line_split[1].strip()
                if name_pattern.match(name):
                    try:
                        return int(line_split[0]), name
                    except ValueError:
                        return

    return await match_syscall(cmd=f'ps -Ao "%p#%c" -ww --no-headers --sort={"-" if last_match else ""}pid',
                               match=_match)


async def find_process_by_command(patterns: Set[Pattern], last_match: bool = False) -> Optional[Tuple[int, str]]:
    def _match(line: str) -> Optional[Tuple[int, str]]:
        line_strip = line.strip()

        if line_strip:
            line_split = line_strip.split('#', 1)

            if len(line_split) > 1:
                for p in patterns:
                    cmd = line_split[1].strip()
                    if p.match(cmd):
                        try:
                            return int(line_split[0]), cmd
                        except ValueError:
                            break

    return await match_syscall(cmd=f'ps -Ao "%p#%a" -ww --no-headers --sort={"-" if last_match else ""}pid',
                               match=_match)


async def find_processes_by_command(commands: Set[str], last_match: bool = False) -> Optional[Dict[str, int]]:
    matches = {}

    def _match(line: str) -> Optional[Dict[str, int]]:
        line_strip = line.strip()

        if line_strip:
            line_split = line_strip.split('#', 1)

            if len(line_split) > 1:
                cmd = line_split[1].strip()
                if cmd in commands and cmd not in matches:
                    try:
                        matches[cmd] = int(line_split[0])
                    except ValueError:
                        pass

            if len(matches) == len(commands):
                return matches

    await match_syscall(cmd=f'ps -Ao "%p#%a" -ww --no-headers --sort={"-" if last_match else ""}pid', match=_match)
    return matches if matches else None


async def find_pids_by_names(names: Collection[str], last_match: bool = False) -> Optional[Dict[str, int]]:
    matches: Dict[str, int] = {}

    def _match(line: str) -> Optional[Dict[str, int]]:
        line_strip = line.strip()

        if line_strip:
            line_split = line_strip.split('#', 1)

            if len(line_split) > 1:
                current_name = line_split[1].strip()

                if current_name and current_name not in matches and current_name in names:
                    try:
                        matches[current_name] = int(line_split[0])
                    except ValueError:
                        pass

        if len(matches) == len(names):
            return matches

    await match_syscall(cmd=f'ps -Ao "%p#%c" -ww --no-headers --sort={"-" if last_match else ""}pid', match=_match)
    return matches if matches else None


async def find_commands_by_pids(pids: Set[int]) -> Optional[Dict[int, str]]:
    if pids:
        matches: Dict[int, str] = {}

        def _match(line: str) -> Optional[Dict[int, str]]:
            line_strip = line.strip()

            if line_strip:
                line_split = line_strip.split('#', 1)

                if len(line_split) > 1:
                    try:
                        current_pid = int(line_split[0])
                    except ValueError:
                        return

                    if current_pid in pids:
                        matches[current_pid] = line_split[1].strip()

            if len(matches) == len(pids):
                return matches

        await match_syscall(cmd='ps -Ao "%p#%a" -ww --no-headers', match=_match)
        return matches if matches else None


def read_current_pids() -> Set[int]:
    return {int(pid) for pid in os.listdir('/proc') if pid.isnumeric()}


async def map_pids_by_ppid() -> Optional[Dict[int, Set[int]]]:
    code, output = await async_syscall('ps -Ao "%P#%p" -ww --no-headers')

    if code == 0 and output:
        all_procs = {}
        for line in output.split('\n'):
            if line:
                line_strip = line.strip()

                if line_strip:
                    line_split = line_strip.split('#', 1)

                    if len(line_split) == 2:
                        try:
                            current_ppid, current_pid = int(line_split[0].strip()), int(line_split[1].strip())
                        except ValueError:
                            continue

                        current_children = all_procs.get(current_ppid, set())
                        current_children.add(current_pid)
                        all_procs[current_ppid] = current_children

        return all_procs


async def find_children(ppids: Set[int], ppid_map: Optional[Dict[int, Set[int]]] = None, children: Optional[List[int]] = None) -> Optional[List[int]]:
    pids_by_ppids = ppid_map if ppid_map is not None else (await map_pids_by_ppid())
    children_list = children if children is not None else []

    if pids_by_ppids:
        current_children = set()
        for pid in ppids:
            pid_children = pids_by_ppids.get(pid)

            if pid_children:
                current_children.update((p for p in pid_children if p not in children_list and p not in ppids))

        if current_children:
            await find_children(current_children, pids_by_ppids, children_list)
            children_list.extend(current_children)

    return children_list


async def run_async_process(cmd: str, user_id: Optional[int] = None, custom_env: Optional[dict] = None,
                            forbidden_env_vars: Optional[Set[str]] = BAD_USER_ENV_VARS,
                            wait: bool = True, timeout: Optional[float] = None, output: bool = True,
                            exception_output: bool = True) \
        -> Tuple[Optional[int], Optional[int], Optional[str]]:
    """
    Runs a process using the async API
    Args:
        cmd:
        user_id: if the process should be executed in behalf of a different user
        custom_env: custom environment variables available for the process to be executed
        forbidden_env_vars: environment variables that should not be passed to the process
        wait: if the process should be waited
        timeout: in seconds
        output: if the process output should be read and returned
        exception_output: if the traceback of an unexpected raised exception should be returned as the output
    Returns: a tuple containing the process id, exitcode and output as a String

    """
    args = {"cmd": cmd, "stdin": subprocess.DEVNULL,
            "stdout": subprocess.PIPE if output else subprocess.DEVNULL,
            "stderr": subprocess.STDOUT if output else subprocess.DEVNULL}

    if user_id is not None:
        args["preexec_fn"] = lambda: os.setuid(user_id)

    if custom_env:
        if forbidden_env_vars:
            args['env'] = {k: v for k, v in custom_env.items() if k not in forbidden_env_vars}
        else:
            args['env'] = custom_env

    try:
        p = await asyncio.create_subprocess_shell(**args)
    except Exception:
        return None, 1, (traceback.format_exc().replace('\n', ' ') if exception_output else None)

    if user_id is not None:  # set default niceness in case the process is executed in behalf of another user
        try:
            os.setpriority(os.PRIO_PROCESS, p.pid, 0)  # always launch a command with nice 0
        except Exception:
            pass  # do nothing in case the priority could not be changed

    should_wait = wait or (timeout and timeout > 0)
    try:
        if should_wait:
            if timeout is None or timeout < 0:
                return p.pid, await p.wait(), ((await p.stdout.read()).decode() if output else None)
            elif timeout and timeout > 0:
                timeout_at = datetime.now() + timedelta(seconds=timeout)

                while datetime.now() < timeout_at:
                    if p.returncode is not None:
                        return p.pid, p.returncode, ((await p.stdout.read()).decode() if output else None)

                    await asyncio.sleep(0.0005)

                raise ProcessTimedOutError(p.pid)

        return p.pid, p.returncode, None
    except ProcessTimedOutError:
        raise
    except Exception:
        return p.pid, 1, (traceback.format_exc().replace('\n', ' ') if exception_output else None)


async def map_processes_by_parent() -> Dict[int, Set[Tuple[int, str]]]:
    exitcode, output = await async_syscall(f'ps -Ao "%P#%p#%c" -ww --no-headers')

    if exitcode == 0 and output:
        proc_tree = dict()

        for line in output.split("\n"):
            line_strip = line.strip()

            if line_strip:
                line_split = line_strip.split('#', 2)

                if len(line_split) > 2:
                    ppid, pid, comm, = (e.strip() for e in line_split)
                    try:
                        ppid, pid = int(ppid), int(pid)
                    except ValueError:
                        continue

                    children = proc_tree.get(ppid)

                    if not children:
                        children = set()
                        proc_tree[ppid] = children

                    children.add((pid, comm))

        return proc_tree


def find_process_children(ppid: int, processes_by_parent: Dict[int, Set[Tuple[int, str]]],
                          comm_to_ignore: Optional[Set[str]] = None, already_found: Optional[Set[int]] = None,
                          recursive: bool = True) \
        -> Generator[Tuple[int, str, int], None, None]:
    found = already_found if already_found is not None else set()

    children = processes_by_parent.get(ppid)

    if children:
        for pid, comm in children:
            if (not comm_to_ignore or comm not in comm_to_ignore) and "<defunct>" not in comm:
                if pid not in found:
                    yield pid, comm, ppid
                    found.add(pid)

                if recursive:
                    for pid_, comm_, ppid_ in find_process_children(ppid=pid, processes_by_parent=processes_by_parent,
                                                                    comm_to_ignore=comm_to_ignore, already_found=found):
                        yield pid_, comm_, ppid_



import asyncio
import re
from abc import ABC, abstractmethod
from datetime import datetime, timedelta
from logging import Logger
from typing import Dict, Optional, Tuple, Generator, AsyncGenerator, Pattern, Set, List

import aiofiles

from guapow import __app_name__
from guapow.common import util
from guapow.common.dto import OptimizationRequest
from guapow.common.model import CustomEnum
from guapow.common.system import async_syscall, map_processes_by_parent, find_process_children
from guapow.common.users import is_root_user
from guapow.service.optimizer.profile import OptimizationProfile

DELIMITER = '%'


class LauncherSearchMode(CustomEnum):
    NAME = 'n'
    COMMAND = 'c'

    @classmethod
    def map_string(cls, string: str) -> "LauncherSearchMode":
        return LauncherSearchMode.COMMAND if string.startswith('/') else LauncherSearchMode.NAME


def gen_possible_launchers_file_paths(user_id: int, user_name: str) -> Generator[str, None, None]:
    if not is_root_user(user_id):
        yield f'/home/{user_name}/.config/{__app_name__}/launchers'

    yield f'/etc/{__app_name__}/launchers'


def map_target(string: str, mapping: str, logger: Logger) -> Optional[Tuple[str, LauncherSearchMode]]:
    string_split = string.split(DELIMITER, 1)

    if len(string_split) > 1:
        mode_str = string_split[0].strip()
        mode = LauncherSearchMode.from_value(mode_str.lower())

        if not mode:
            mode = LauncherSearchMode.map_string(string)
            logger.warning(f"Invalid launcher target '{mode_str}' for mapping: {mapping}. Default type '{mode.value}' will be considered")

        return string_split[1].strip(), mode
    else:
        return string, LauncherSearchMode.map_string(string)


def map_launchers_dict(launchers: dict, logger: Logger) -> Optional[Dict[str, Tuple[str, LauncherSearchMode]]]:
    if launchers:
        res = {}
        for launcher, target in launchers.items():
            launcher_strip = launcher.strip()

            if not launcher_strip:
                continue

            target_strip = target.strip()

            if not target_strip:
                continue

            res[launcher_strip] = map_target(target_strip, f'{launcher}{DELIMITER}{target}', logger)

        return res if res else None


async def map_launchers_file(wrapper_file: str, logger: Logger) -> Optional[Dict[str, Tuple[str, LauncherSearchMode]]]:
    wrappers = {}

    async with aiofiles.open(wrapper_file) as f:
        async for line in f:
            line_strip = line.strip()

            if line_strip and not line_strip.startswith('#'):
                line_split = line_strip.split('#', 1)[0].split('=', 1)

                if len(line_split) == 2:
                    key = line_split[0].strip()

                    if key:
                        val = line_split[1].strip()

                        if val:
                            wrappers[key] = map_target(val, line_strip, logger)

    return wrappers


class LauncherMapper(ABC):
    """
    Responsible for mapping the real processes to be optimized since a source process.
    """

    def __init__(self, check_time: float, found_check_time: float, logger: Logger):
        """
        Args:
            check_time: the maximum amount of time the mapper should be looking for a match (seconds)
            found_check_time:
                the maximum amount of time the mapper should be still looking for matches after some already found
                (seconds)
            logger:
        """
        self._log = logger
        self._check_time = check_time
        self._found_check_time = found_check_time

    @abstractmethod
    async def map_pids(self, request: OptimizationRequest, profile: OptimizationProfile) -> AsyncGenerator[int, None]:
        pass


class ExplicitLauncherMapper(LauncherMapper):
    """
    For mappings declared in the 'launchers' file
    """

    def __init__(self, check_time: float, found_check_time: float, logger: Logger,
                 iteration_sleep_time: float = 0.1):
        super(ExplicitLauncherMapper, self).__init__(check_time=check_time,
                                                     found_check_time=found_check_time,
                                                     logger=logger)
        self._iteration_sleep_time = iteration_sleep_time

    @staticmethod
    async def map_process_by_pid(mode: LauncherSearchMode, ignore: Set[int]) -> Optional[Dict[int, str]]:
        if mode:
            mode_str = "a" if mode == LauncherSearchMode.COMMAND else "c"
            exitcode, output = await async_syscall(f'ps -Ao "%p#%{mode_str}" -ww --no-headers')

            if exitcode == 0 and output:
                pid_comm = dict()

                for line in output.split("\n"):
                    line_strip = line.strip()

                    if line_strip:
                        line_split = line_strip.split('#', 1)

                        if len(line_split) > 1:
                            try:
                                pid = int(line_split[0])
                            except ValueError:
                                continue

                            if pid in ignore:
                                continue

                            pid_comm[pid] = line_split[1].strip()

                return pid_comm

    async def find_wrapped_process(self, wrapped_target: Tuple[str, LauncherSearchMode], launcher: str,
                                   source_pid: int) -> AsyncGenerator[int, None]:
        wrapped_name, search_mode = wrapped_target[0].strip(), wrapped_target[1]
        wrapped_regex = util.map_any_regex(wrapped_name)
        self._log.debug(f"Looking for mapped process with {search_mode.name.lower()} '{wrapped_name}' "
                        f"(launcher={launcher})")

        latest_found_timeout = None
        found = set()
        time_init = datetime.now()
        timeout = time_init + timedelta(seconds=self._check_time)
        while datetime.now() < timeout:
            if latest_found_timeout and datetime.now() >= latest_found_timeout:
                self._log.debug(f"Launcher mapping search timed out earlier (source_pid={source_pid})")
                return

            pid_process = await self.map_process_by_pid(search_mode, ignore=found)

            if pid_process:
                for pid, command in pid_process.items():
                    if wrapped_regex.match(command):
                        if self._found_check_time >= 0:
                            latest_found_timeout = datetime.now() + timedelta(seconds=self._found_check_time)

                        self._log.info(f"Mapped process '{command}' ({pid}) found")
                        yield pid
                        found.add(pid)

            if self._iteration_sleep_time > 0:
                await asyncio.sleep(self._iteration_sleep_time)

        if not found:
            timeout_secs = (datetime.now() - time_init).total_seconds()
            self._log.warning(f"Could not find process with {search_mode.name.lower()} '{wrapped_name}' "
                              f"(launcher={launcher}, source_pid={source_pid}). "
                              f"Timed out in {timeout_secs:.2f} seconds")

    async def map_pids(self, request: OptimizationRequest, profile: OptimizationProfile) -> AsyncGenerator[int, None]:
        if profile.launcher and profile.launcher.skip_mapping:
            self._log.info(f"Skipping launcher mapping for {profile.get_log_str()} (pid: {request.pid})")
            return

        if profile.launcher and profile.launcher.mapping:
            self._log.debug(f"Checking mapped launchers for {profile.get_log_str()} (pid: {request.pid})")
            launchers = map_launchers_dict(launchers=profile.launcher.mapping, logger=self._log)
        else:
            launchers = None

            for file_path in gen_possible_launchers_file_paths(request.user_id, request.user_name):
                try:
                    self._log.debug(f"Checking mapped launchers on '{file_path}' (request: {request.pid})")
                    launchers = await map_launchers_file(file_path, self._log)
                    break
                except FileNotFoundError:
                    self._log.debug(f"Launchers file '{file_path}' not found (request: {request.pid})")

        if launchers:
            file_name = request.command.split('/')[-1].strip()

            wrapped_target = launchers.get(file_name)

            if not wrapped_target:  # if there is no exact match, check any name with regex
                for name, real_name in launchers.items():
                    if '*' in name:
                        if util.map_any_regex(name).match(file_name):
                            wrapped_target = real_name
                            break

            if wrapped_target:
                async for pid in self.find_wrapped_process(wrapped_target, file_name, request.pid):
                    yield pid

        else:
            self._log.debug("No valid launchers mapped found")


class SteamLauncherMapper(LauncherMapper):

    def __init__(self, check_time: float, found_check_time: float, logger: Logger,
                 iteration_sleep_time: float = 0.1):
        super(SteamLauncherMapper, self).__init__(check_time=check_time,
                                                  found_check_time=found_check_time,
                                                  logger=logger)
        self._re_steam_cmd: Optional[Pattern] = None
        self._to_ignore: Optional[Set[str]] = None  # processes that should not be optimized
        self._re_proton_command: Optional[Pattern] = None
        self._iteration_sleep_time = iteration_sleep_time  # used to avoid CPU overloading while looking for targets

    @property
    def re_steam_cmd(self) -> Pattern:
        if not self._re_steam_cmd:
            self._re_steam_cmd = re.compile(r'^.+/(\w+)\s+SteamLaunch\s+.+', re.IGNORECASE)

        return self._re_steam_cmd

    @property
    def re_proton_command(self) -> Pattern:
        if not self._re_proton_command:
            self._re_proton_command = re.compile(r'^.+/proton\s+waitforexitandrun\s+.+$', re.IGNORECASE)

        return self._re_proton_command

    @property
    def to_ignore(self) -> Set[str]:
        if self._to_ignore is None:
            self._to_ignore = {"wineserver", "services.exe", "winedevice.exe", "plugplay.exe", "svchost.exe",
                               "explorer.exe", "rpcss.exe", "tabtip.exe", "wine", "wine64", "wineboot.exe",
                               "cmd.exe", "conhost.exe", "start.exe", "steam-runtime-l", "proton", "gzip",
                               "steam.exe", "python", "python3", "OriginWebHelper", "Origin.exe",
                               "OriginClientSer", "QtWebEngineProc", "EASteamProxy.ex", "ActivationUI.ex",
                               "EALink.exe", "OriginLegacyCLI", "IGOProxy.exe", "IGOProxy64.exe", "igoproxy64.exe",
                               "ldconfig", "UPlayBrowser.exe", "UbisoftGameLaun", "upc.exe", "UplayService.ex",
                               "UplayWebCore.ex", "CrRendererMain", "regsvr32", "CrGpuMain", "CrUtilityMain",
                               "whql:off", "PnkBstrA.exe"}

        return self._to_ignore

    def find_target_in_hierarchy(self, reverse_hierarchy: List[str], root_element_pid: int,
                                 processes_by_parent: Optional[Dict[int, Set[Tuple[int, str]]]] = None,
                                 pid_by_comm: Optional[Dict[str, int]] = None) -> Optional[int]:

        if len(reverse_hierarchy) == 1:
            return root_element_pid

        comm_pid = dict() if pid_by_comm is None else pid_by_comm

        if reverse_hierarchy[-1] not in comm_pid:
            comm_pid[reverse_hierarchy[-1]] = root_element_pid

        for idx, comm in enumerate(reverse_hierarchy):
            pid = comm_pid.get(comm)

            if pid is None:
                parent_id = comm_pid.get(reverse_hierarchy[idx + 1])

                if not parent_id:
                    continue  # the iteration must continue if the parent id is not mapped yet

                parent_children = processes_by_parent.get(parent_id)

                if not parent_children:
                    return   # if the parent has no children, it will not be possible to find the current comm's pid

                try:
                    pid = next(pid_ for pid_, comm_ in sorted(parent_children, reverse=True) if comm_ == comm)
                except StopIteration:
                    return  # the current comm could not be found, so stop the iteration

                comm_pid[comm] = pid

            if idx == 0:  # if current element is the target, return it immediately
                return pid

            # restart the find
            return self.find_target_in_hierarchy(reverse_hierarchy=reverse_hierarchy,
                                                 root_element_pid=root_element_pid,
                                                 processes_by_parent=processes_by_parent,
                                                 pid_by_comm=comm_pid)

    def map_expected_hierarchy(self, request: OptimizationRequest, root_comm: Optional[str] = None) -> List[str]:
        hierarchy = list()

        if "/steamapps/common/SteamLinux" in request.command:
            self._log.debug(f"Steam command comes from container (pid: {request.pid})")
            hierarchy.append("pressure-vessel")
            hierarchy.append("pv-bwrap")
        elif self.re_proton_command.match(request.command):
            hierarchy.append("python3")

        if root_comm:
            hierarchy.append(root_comm)

        return hierarchy

    def extract_root_process_name(self, command: str) -> str:
        root_cmd = self.re_steam_cmd.findall(command)

        if root_cmd:
            return root_cmd[0]

    async def map_pids(self, request: OptimizationRequest, profile: OptimizationProfile) -> AsyncGenerator[int, None]:
        if profile.steam:
            steam_root_comm = self.extract_root_process_name(request.command)

            if not steam_root_comm:
                self._log.warning(f'Command not from Steam: {request.command} (pid: {request.pid})')
            else:
                self._log.debug(f'Steam command detected for request (pid: {request.pid})')
                expected_hierarchy = self.map_expected_hierarchy(request, steam_root_comm)
                timeout = datetime.now() + timedelta(seconds=self._check_time)

                latest_found_timeout = None  # timeout for every time a target is found (to stop faster)
                pid_by_comm = dict()  # to save which processes were previously mapped
                already_found: Set[int] = set()  # processes already yielded
                target_ppid = None  # parent with the target children

                to_ignore = {*expected_hierarchy, *self.to_ignore}  # target children to ignore

                while datetime.now() < timeout:
                    if latest_found_timeout and datetime.now() >= latest_found_timeout:
                        self._log.debug(f"Steam subprocesses search timed out earlier (source_pid={request.pid})")
                        return

                    parent_procs = await map_processes_by_parent()

                    if target_ppid is None:
                        target_ppid = self.find_target_in_hierarchy(reverse_hierarchy=expected_hierarchy,
                                                                    root_element_pid=request.pid,
                                                                    processes_by_parent=parent_procs,
                                                                    pid_by_comm=pid_by_comm)
                        if target_ppid is not None:
                            self._log.debug(f"Target Steam process parent found (pid={target_ppid}, "
                                            f"comm={expected_hierarchy[0]}) (source_pid={request.pid})")

                    if target_ppid is not None:
                        for pid_, comm_, ppid_ in find_process_children(ppid=target_ppid,
                                                                        processes_by_parent=parent_procs,
                                                                        already_found=already_found,
                                                                        comm_to_ignore=to_ignore,
                                                                        recursive=False):
                            if self._found_check_time >= 0:
                                latest_found_timeout = datetime.now() + timedelta(seconds=self._found_check_time)

                            self._log.info(f"Steam child process found: {comm_} (pid={pid_}, ppid={ppid_})")
                            yield pid_

                    if self._iteration_sleep_time > 0:
                        await asyncio.sleep(self._iteration_sleep_time)

                self._log.debug(f"Steam subprocesses search timed out (source_pid={request.pid})")


class LauncherMapperManager(LauncherMapper):
    def __init__(self, check_time: float, found_check_time: float,
                 logger: Logger, mappers: Optional[Tuple[LauncherMapper, ...]] = None):
        super(LauncherMapperManager, self).__init__(check_time, found_check_time, logger)

        if mappers:
            self._sub_mappers = mappers
        else:
            sub_classes = LauncherMapper.__subclasses__()
            self._sub_mappers = tuple(cls(check_time, found_check_time, logger) for cls in sub_classes
                                      if cls != self.__class__)

    async def map_pids(self, request: OptimizationRequest, profile: OptimizationProfile) -> AsyncGenerator[int, None]:
        any_mapper_yield = False
        for mapper in self._sub_mappers:
            if any_mapper_yield:  # if any mapper already returned something, stop the iteration
                return

            async for real_id in mapper.map_pids(request, profile):
                if real_id is not None:
                    any_mapper_yield = True
                    yield real_id

    def get_sub_mappers(self) -> Optional[Tuple[LauncherMapper]]:
        if self._sub_mappers is not None:
            return tuple(self._sub_mappers)
